/*!
    \file   df_lua_connection.h
    \author Bernd Ottmann
    \date   21.02.2003
    \brief  Deklarationen für die Funktionen der Geräteverbindungen und deren direktem Zugriff.
*/
#ifndef DF_LUA_CONNECTION_H
#define DF_LUA_CONNECTION_H

#include "df_lua_api.h"

#if (DF_API_VERSION >= 0x04030600)
/*! 
    \brief Oeffnet die Schnittstelle fuer die Kommunikation mit PZE Master IV.

    Initialisiert den Kanal. 
    Je nach angegebenem Typ wird ein anderes Initialisierungsverfahren verwendet 
    um die entsprechend nachfolgenden Funktionsaufrufe der exportierten Funktionen 
    richtig umzusetzen.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Die von der DLL zu verwendende Busnummer. (Multimasteransatz)
    \param connectionType Der zu verwendende Schnittstellentyp.
                 1 -> RS232 direkt oder über Umsetzter (RS232 <-> RS485)
                 2 -> Funk über den Umsetzer (RS232 <-> BIM2) ohne eigenen Kontroller.
                 3 -> TCP / IP - Verbindung.
    \param connectionString Je nach Schnittstellentyp:
                       1,2 -> Zeiger auf Zeichenkette für die Schnittstelle. "COM1" .. "COM4".
                       3 -> Zeiger auf Zeichenkette mit IP - Adresse oder Hostname.
    \param connectionValue Je nach Schnittstellentyp:
                     1,2 -> Übertragungsrate der Schnittstelle. (Standard ist: 38400)
                     3 -> Portnummer des TCP / IP Anschlusses.
    \param connectionTimeout Je nach Schnittstellentyp:
                       1,2 ->Timeout für Lesen von Schnittstelle (Standard ist: 150)
                       3 -> Timeout für Lesen von Socket (Standard ist: 3000)

    \return Kommunikationskanal oeffnen.
    \retval 1 Schnittstelle erfolgreich geoeffnet.
    \retval 0 Fehler bei Initialisierung.
	
	\snippet examples.cpp DFCComOpenIV
*/
int l_open( lua_State *L );

/*! 
    \brief Schliessen des geoeffneten Kanals.

    Je nach dem Typ des Kanals wird entweder die serielle Schnittstelle oder
    der Socket geschlossen.

    \param connectionId Kommunikationskanalnummer.

    \return Keine.
	
	\snippet examples.cpp DFCComClose
*/
int l_close( lua_State *L );

/*! 
    \brief Prueft ob die Schnittstelle bereits Initialisiert oder auch geoeffnet wurde. 

    \param connectionId Kommunikationskanalnummer.

    \return Kommunikationskanal oeffnen.
    \retval -1 Kanalobjekt noch nicht verfügbar, Schnittstelle nicht initialisiert oder geoeffnet.
    \retval 0 Schnittstelle nicht geoeffnet.
    \retval 1 Schnittstelle geoeffnet.
*/
int l_isOpen( lua_State *L );

/*! 
    \brief Schreibt Zeichen in den Kanal.

    Schreiben von Daten in den Kanal.

    \param connectionId Kommunikationskanalnummer.
    \param buffer Zeiger auf Daten.
    \param toWrite Laenge der zu schreiben Daten.
    \param written Anzahl der geschriebenen Daten.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.

    \return Ergebnis der Ausfuehrung.
    \retval 0 Funktion konnte nicht ausgefuehrt werden.
    \retval 1 Funktion wurde erfolgreich ausgefuehrt.

    \version Ab DLL - Version 02.00.18 enthalten.
*/
int l_write( lua_State *L );

/*! 
    \brief Liest Zeichen vom Kanal.

    Lesen von Daten aus dem Kanal.

    \param connectionId Kommunikationskanalnummer.
    \param buffer Zeiger auf Daten.
    \param toRead Laenge der zu lesenden Daten.
    \param read Anzahl der gelesenen Daten.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.

    \return Ergebnis der Ausfuehrung.
    \retval 0 Funktion konnte nicht ausgefuehrt werden.
    \retval 1 Funktion wurde erfolgreich ausgefuehrt.

    \version Ab DLL - Version 02.00.18 enthalten.
*/
int l_read( lua_State *L );

#endif

#endif
