/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#ifndef CHANNEL4_PROTOCOL_H
#define CHANNEL4_PROTOCOL_H

// Basisklasse einbinden.
#include "Channel4Base.h"
#include <time.h>
#include <string>

#include "../Common/crc.h"
#include "../Common/aes.h"
#include "TimeboyStation.h"
#include "../Common/aescrypt.h"

///////////////////////////////////////
// Defines zum Packet

// Makros fuer den Zugriff auf die Packetdaten
#define IV_PACKET_START(x) x[0]
#define IV_PACKET_DEVICEID(x) x[1]
#define IV_PACKET_DEVICEID_REMOTE(x) x[2]
#define IV_PACKET_SIZE_CONTROLBLOCK(x) x[3]
#define IV_PACKET_SIZE(x) (*(unsigned short*)&x[4])
#define IV_PACKET_FLAGS(x) x[6]
#define IV_PACKET_RESERVE(x) x[7]
#define IV_PACKET_CRC_HEAD(x) (*(unsigned short*)&x[8])

#define IV_PACKET_COMMAND(x) x[IV_PACKET_LENGTH_HEAD]
#define IV_PACKET_MODULE(x) (*(unsigned short*)&x[IV_PACKET_LENGTH_HEAD + IV_PACKET_LENGTH_CONTROLBLOCK + sizeof(unsigned char)])
#define IV_PACKET_FUNCTION(x) (*(unsigned short*)&x[IV_PACKET_LENGTH_HEAD + IV_PACKET_LENGTH_CONTROLBLOCK + sizeof(unsigned char) + sizeof(unsigned short)])

#define IV_PACKET_CRC(x) (*(unsigned short*)&x[IV_PACKET_LENGTH(x) - sizeof(unsigned short)])

// Grundangaben
#define IV_PACKET_LENGTH_MAX                0x300 // Maximalgroesse fuer ein Paketspeicher
#define IV_PACKET_LENGTH_HEAD               10    // Laenge des Packetkopfes
#define IV_PACKET_LENGTH_CONTROLBLOCK        1    // Laenge der Steuerinformation
#define IV_PACKET_LENGTH_FRAME              12    // Minimalframegroesse (Paketkopf + 16bit CRC)
#define IV_PACKET_LENGTH_DATA               (IV_PACKET_LENGTH_MAX - IV_PACKET_LENGTH_FRAME - IV_PACKET_LENGTH_CONTROLBLOCK)

// Paketsonderzeichen 
#define IV_PACKET_SIGN_START                0x7E  // ~ als Startzeichen der Pakete Byte 1
#define IV_PACKET_SIGN_ESCAPE               0x7F  // als Escapezeichen der Pakete

// Paketflags Byte 6
// Achtung: !!! Die unteren 3 bit geben einen Wert an und sind somit keine seperaten Flags!!!
#define IV_PACKET_CRYPT_MODE_NA             0x00 // Not available. Keine Unterstuetzung von verschluesselter Kommunikation.
#define IV_PACKET_CRYPT_MODE_NO             0x01 // No. Verschluesselte Kommunikation wird unterstuetzt muss aber unverschluesselt kommunizieren.
#define IV_PACKET_CRYPT_MODE_DEFAULT        0x02 // Default. Verschluesselung mit dem Datafox-Standard-Schluessel.
#define IV_PACKET_CRYPT_MODE_PASSPHRASE     0x03 // Passphrase. Es muss mit einem Benutzerdefinierten Passwort verschluesselt werden.
// Dieses hier sind nun Flags.
#define IV_PACKET_FLAG_CRYPTED              0x04 // Dateninhalt liegt verschluesselt vor.
#define IV_PACKET_FLAG_RESERVED1            0x08 // 
#define IV_PACKET_FLAG_FLASH_SERVICE        0x10 // Flash Service: Globaler Wear-Level ist >= 6				
#define IV_PACKET_FLAG_FLASH_WARNING        0x20 // Flash Warning: CRC Fehler im Ringspeicher oder Listenbereich. (Setup und Listen neu übertragen)		
#define IV_PACKET_FLAG_FLASH_LOCKED         0x40 // Flash Error: Keine freien Seiten mehr verfügbar, Gerät befindet sich im Bedienstopp.
#define IV_PACKET_FLAG_RESERVED2            0x80 // 
///////////////////
// Paketdaten

// Paketkommando (unteres Nibble)
#define IV_PACKET_VERSION_4                 0x04 // Packetversion 4 seit FW 04.01.xx und Bootloader 04.xx
// Paketkommando (oberes Nibble)
#define IV_PACKET_ANSWER                    0x80 // Packet ist eine Antwort
#define IV_PACKET_REQUEST                   0x40 // Packet ist eine Anfrage
#define IV_PACKET_ERROR                     0x20 // Packet transportiert einen Fehler (Anfrage oder Antwort)


/////////////////////////////
// Werte fuer die Umgebung

// Standardwerte als Vorgabe
#define IV_DEFAULT_DEVICEID                 0x0000 // Modbusadresse fuer Library
#define IV_DEFAULT_DEVICEID_STANDALONE      0x00FE // Modbusadresse fuer Direktverbindung

#define IV_PENDING_TIMEOUT                  60000 // Timeout fuer eine Aktive Antwort (ab 4.1.5.9 von 15 auf 25s, ab 4.2.3 auf 60s)
#define IV_PENDING_FINGERPRINT_TIMEOUT      60000 // Timeout fuer eine Aktive Antwort (ab 4.2.1, ab 4.2.3 von 50 auf 60s)
#define IV_PENDING_RESET_TIMEOUT            95000 // Timeout fuer eine Aktive Antwort bei SYSTEM.SYSTEMRESET (ab 4.2.1, ab 4.3.3 von 5000 auf 95s)
#define IV_PENDING_TRIALS(to)               (!(IV_PENDING_TIMEOUT % to) ? IV_PENDING_TIMEOUT / to : IV_PENDING_TIMEOUT / to + 1)
#define IV_PENDING_FINGERPRINT_TRIALS(to)   (!(IV_PENDING_FINGERPRINT_TIMEOUT % to) ? IV_PENDING_FINGERPRINT_TIMEOUT / to : IV_PENDING_FINGERPRINT_TIMEOUT / to + 1)
#define IV_PENDING_RESET_TRIALS(to)         (!(IV_PENDING_RESET_TIMEOUT % to) ? IV_PENDING_RESET_TIMEOUT / to : IV_PENDING_RESET_TIMEOUT / to + 1)
#define IV_SKIPCOUNT                        1024 // Maximale Anzahl Zeichen die auserhalbt des Packetzusammenhangs zu lesen sind.

// Contexte fuer call-Routine
#define IV_CONTEXT_MASTER                    0x0000 // Ausfuehrung als Kommunikationsbefehl
#define IV_CONTEXT_SLAVE                     0x0001 // Pruefen auf anstehende Kommunikationsbefehle der Gegenstelle
#define IV_CONTEXT_TRANSMIT_ANSWER           0x0002 // Senden einer Antwort
#define IV_CONTEXT_TRANSMIT_ANSWER_ERROR     0x0004 // Senden einer Antwort als Fehler

// Maximale Laenge eines Fehlertextes
#define ERROR_TEXT_SIZE                      255

// Auswertung der Fehlerbestandteile
#define IV_RESULT_MODULE(res)                ((res >> 8) & 0xFF)
#define IV_RESULT_FUNCTION(res)              (res & 0xFF)


// Fuer die Verwendung des Kanals als ein Aktiver
enum ActiveState { 
    is_free = 0,      // Kanal ist nicht in Verwendung
    is_accepting, // Kanal nimmt eine Verbindung entgegen
    is_usable,    // Kanal hat die Verbindung angenommen und ermittelt Daten vom Gerät.
    is_binding,   // Kanal durchläuft die Kanalbindungsprüfung.
    is_active     // Kanal ist Betriebsbereit.
};

/////////////////////////////
// Error's
#include "../Common/AEErrorDef.h"

///////////////////////////////////////
// Modul und Funktionsbasis

/** In diesem Module sind alle Kommunikationskontroll Funktionen enthalten.
*
* Es handelt sich hierbei um ein Modul zur Kontrolle der auszufuehrenden Funktionen
* der abgeleiteten Module.
*
*/
#define IV_MODULE_BASE                       ((short)0xFFFF)
// Undefinierte Modulfunktion
#define IV_FUNCTION_BASE                     ((short)0xFFFF)

/** In diesem Module sind alle Protokolspezifischen Funktionen enthalten.
*
* 
*
*/
#define IV_MODULE_PROTOCOL                       0x0000

// Prueft die Erreichbarkeit eines Geraets. Es wird nur ein einzelnes Paket versendet.
#define IV_FUNCTION_PING                         0x0000


/// Basisklasse fuer pure protokollbasierte Kommunikationsaufgaben.
class Channel4Protocol : public Channel4Base
{
    friend class ChannelBridge;
public:
    Channel4Protocol( short channelId = IV_DEFAULT_CHANNELID, short deviceId = IV_DEFAULT_DEVICEID );
    virtual ~Channel4Protocol();

public:
    /// Oeffnen der Schnittstelle
    int open( enumType type, const std::string &string, int val, int timeout )
    {
        // Bei Autobaud und keiner Vorgabe (val == 0) wird die zuvor gewaehtle Baudrate oder 38400 in ChannelBase verwendet.
        return Channel4Base::open( type, string, val, timeout );
    }
    /// Schliessen der Schnittstelle.
    virtual int close()
    {
        return Channel4Base::close();
    }

    // Autobauding mit prio 38400, 9600, 19200. ( HK hinzugefügt 57600, 115200 )
    bool getAutobaud()
    {
        return _autobaud;
    }
    void setAutobaud(bool val)
    {
        _autobaud = val;
    }

    // Speicher fuer Pakete/Daten
public:
    unsigned char _packetWrite[IV_PACKET_LENGTH_MAX];
    //unsigned char m_lastRequest[IV_PACKET_LENGTH_MAX];
    unsigned char _packetRead[IV_PACKET_LENGTH_MAX];
    unsigned char _packetData[IV_PACKET_LENGTH_DATA]; // 21 Byte Paketframe
    unsigned char _tmp[IV_PACKET_LENGTH_MAX * 2];
    unsigned char *_tmpPos;

    void reset();

private:
    int _packetStart;
    int _skipCount;
    unsigned char _packetNumber; // Paketzaehler
    bool _autobaud;
protected:
    unsigned char _lastFlags;    // Halten der letzten empfangenen Flagvariable aus Packet. Für Flash-Service Flags etc...
private:
    double _qs;

    clock_t _clockWrite;    // Dauer des letzten Versendevorgangs
    clock_t _clockRead;     // Dauer des letzten Empfangvorgangs
    clock_t _clocksWrite;   // Gesamtdauer der Versendevorgaenge
    clock_t _clocksRead;    // Gesamtdauer der Empfangsvorgaenge
    int _packetWriteCounter;  // Anzahl gueltig versendeter Pakete
    int _packetWritePayloadSize; // Anzahl der versendeten Nutzdaten-Bytes
    int _packetReadCounter;     // Anzahl gueltig empfangener Pakete
    int _packetLostCounter;// Anzahl unbeantworteter Pakete
    int _packetReadPayloadSize;    // Anzahl der gelesenen Nutzdaten-Bytes

    short _deviceId; // Die zu verwendende DeviceID der Klasse
    short _deviceIdRemote; // Die zu verwendete DeviceID des Remote der Klasse
protected:
    // Membervariablen fuer call-Routine
    short _previousModule;   // Zuletzt ausgefuehrtes Modul
    short _previousFunction; // Zuletzt ausgefuehrte Funktion
    int _result;         // Returnwert der call-Routine

    // Routinen fuer DeviceID's
public:
    short getDeviceId() 
    { 
        return _deviceId; 
    }
    short setDeviceId( short deviceId )
    { 
        short deviceIDPrev = _deviceId; 
        _deviceId = deviceId; 
        return deviceIDPrev; 
    }
    short getDeviceIdRemote() 
    { 
        return _deviceIdRemote; 
    }
    short setDeviceIdRemote( short deviceIdRemote )
    { 
        short deviceIDPrev = _deviceIdRemote; 
        _deviceIdRemote = deviceIdRemote; 
        return deviceIDPrev; 
    }

    // Routinen zur Ermittlung von sonstigen Memberwerten
public:
    inline unsigned char getPacketNumber()
    {
        return _packetNumber;
    }

    // Standardroutinen
protected:
    // Zeiger auf CallBack-Funktionen
    int (*_callBackFunction)();
    int (*_callBackFunctionWithParams)( unsigned int size, unsigned int count );
public:
    void setCallBackFunction( int (*pFunction)() );
    void setCallBackFunctionWithParams( int (*pFunction)( unsigned int, unsigned int ) );

    // Routinen fuer CRC 16 sowie 32bit
private:
    unsigned int _crc32_table[256]; // Lookup table array
    void initCrc32Table();
    unsigned int reflect( unsigned int ref, char ch );
public:
    unsigned int crc32( unsigned char *data, unsigned int size, unsigned int crc = 0xFFFFFFFF, unsigned char closure = TRUE );
    unsigned short crc16( unsigned char *data, unsigned int size, unsigned short crc = 0xFFFF );

    // Routinen fuer Verschluesselung der Paketdaten
private:
    unsigned char _sbox[256];
    unsigned char _k[256];
    unsigned char _key[32];
protected:
    // Verschluesselungsroutinen
    void df4LoadCryptKey();
    void df4Init( int binaryKey );
    void df4( unsigned char *data, int size );
    int isKeyOEM2()
    { 
        return FALSE;
    }

protected:
    time_t _timeReceivePacket;  // Zeitpunkt des letzten Packetempfangs
    time_t _timeTransmitPacket; // Zeitpunkt des letzten Packetversands
    // Packetdaten ggf. verschluesseln.
    virtual int writeProtocolCrypt()
    { 
        return IV_ERROR_SUCCESS; 
    }
    // Packetdaten Byteweise schreiben und dabei Maskieren, komplettpacket senden.
    int  writeProtocol( unsigned char *buffer, int writeTrials = IV_WRITE_TRIALS );
    // Aufbereiten und senden eines Protokollpakets
    int writePacket( short deviceIdRemote, unsigned char type, short module, short func, unsigned char *buf, int count, int writeTrials = IV_WRITE_TRIALS );
    // Zu versendende Daten ggf. verschluesseln.
    virtual int writeStreamCrypt( std::string &buffer )
    { 
        (void)buffer; 
        return IV_ERROR_SUCCESS;
    }
    // Zu empfangende Daten ggf. entschluesseln.
    virtual int readStreamCrypt( std::string &buffer )
    { 
        (void)buffer; 
        return IV_ERROR_SUCCESS;
    }
    // Ermitteln ob akutell eine Verschluesselungssession aktiv ist.
    virtual bool isCrypt()
    { 
        return false;
    }

    // Packetdaten Byteweise empfangen und Demaskierung durchfuehren.
    int readProtocol( unsigned char *buffer, int readTrials = IV_READ_TRIALS );
    int readTimeboyStation( unsigned char *buffer, int readTrials = IV_READ_TRIALS );
    // Packetdaten ggf. entschluesseln
    virtual int readProtocolCrypt()
    { 
        return IV_ERROR_SUCCESS;
    }
    // Empfang und aufbereiten eines Protokollpakets
    int readPacket( short deviceId, unsigned char &type, short &module, short &func, unsigned char *buf, int &count, int readTrials = IV_READ_TRIALS );

    // Daten eines Packets raustracen
    void tracePacket( unsigned char *packetBuf, bool crlf = true );

    //#################################
    // CallRoutinen fuer die Abarbeitung anstehender Kommunikationsbefehle
    virtual int remoteCall( short module, short func, unsigned char *data, int &count );

public:
    // Einfache Anfrage ob das Geraet antwortet.
    int ping( short deviceId, unsigned char *data, int *count );

public:
    // Verwaltungsklasse fuer Dockingstation im Channel
    TimeboyStation _timeboyStation;
    // Ermittlung ob es sich um einen Slot einer TimeboyStation handelt.
    int isTimeboySlot( short deviceIdRemote )
    {
        if ( _timeboyStationEnabled )
            return _timeboyStation.isSlot( (unsigned char)deviceIdRemote );

        return 0;
    }
    // Beginn und Ende der Kommunikation ueber einen TimeboySlot
    virtual int setTimeboySlotState( short deviceId, int state, int commandResult = IV_ERROR_SUCCESS )
    {
        if ( _timeboyStationEnabled )
            return _timeboyStation.setTimeboySlotState( deviceId, state, commandResult );

        return -IV_ERROR_SLOT_NONE;
    }
    void resetTimeboyStation()
    {
        if ( _timeboyStationEnabled )
            _timeboyStation.resetSlots();
    }
    void setTimeboyStationEnabled( bool enabled )
    {
        if ( _timeboyStationEnabled )
            _timeboyStationEnabled = enabled;
    }
protected:
    unsigned int _timeboyBaseAddress;
    bool _timeboyStationEnabled;
 
public:
    // Ausfuehren eines Protokollbefehls
    virtual int call( short context, short module, short func, unsigned char *buf, int &count, int contextValue = -1 );

private:
    volatile ActiveState _activeState;

public:
    // Ist das Kanalobjekt in Verwendung als Aktive Verbindung.
    inline void setActiveState( ActiveState as ) 
    { 
        _activeState = as; 
    }
    inline ActiveState getActiveState() 
    { 
        return _activeState; 
    }

    // Fehlerbearbeitung
public:
    void resetError();
    inline int getLastResult()
    { 
        return _result;
    }
    // Fehlertexte fuer die Kommunikation/Module/Funktionen
    short _errorCommunicationNum;
    std::string _errorCommunication;
    void getErrorCommunication( short error, std::string &text );

    short _errorModuleNum;
    std::string _errorModule;
    virtual void getErrorModule( short module, short error, std::string &text );

    short _errorFunctionNum;
    std::string _errorFunction;
    virtual void getErrorFunction( short module, short func, short error, std::string &text );

    void traceError( short deviceID, short module, short func );

protected:
    int callServer();

    volatile int _endSignalServer;
    volatile int _initSignalServer;

    volatile enum threadCommandServer {
        commandIdle,
        commandSuspending,
        commandSuspend,
        commandResuming,
        commandResume
    } _threadCommandServer;
    volatile int _suspendCounterServer;
    unsigned char _serviceStateServer;
};

#endif
