#ifndef FILEINTELHEX_H
#define FILEINTELHEX_H

// Grunddaten und Grundoffsets
#define INTELHEX_DATA_COUNT       20  // Anzahl Datenbyte pro Datensatz
#define INTELHEX_START_SIGN       ':' // Startzeichen eines Datensatzes
#define INTELHEX_START            0  // start of record
#define INTELHEX_COUNT            1  // byte count
#define INTELHEX_ADDRESS          3  // load address
#define INTELHEX_TYPE             7  // record type
#define INTELHEX_DATA             9  // record data
#define INTELHEX_CHECKSUM         7  // record checksum

// Groessen fuer Textpuffer der Intelhex-Records
#define INTELHEX_LENGTH_HEAD      9 // Anzahl Zeichen fuer Recordbeginn zu Daten
#define INTELHEX_LENGTH_FRAME     (INTELHEX_LENGTH_HEAD + 2) // Anzahl Zeichen fuer Recordbeginn zu Daten + CRC
#define INTELHEX_LENGTH_MAX       (INTELHEX_LENGTH_FRAME + 510) // Anzahl Zeichen fuer Recordframe + Daten

#include <string>
#include <boost/format.hpp>

#include <vector>

class FileIntelHex
{
public:
	FileIntelHex();
	virtual ~FileIntelHex();

public:
	enum EnumErrorState
	{
		None,                       // Kein Fehler im Zustand.
		FileOpen,                   // Oeffnen der Datei gescheitert.
		// ErrorState der IntelHex-Verarbeitung
		RecordEmpty,                // Die uebergebene Datenzeile ist leer.
		RecordCrc,                  // Der Berechnete CRC stimmt nicht mit dem im Datensatz ueberein.
		RecordFormat,               // Der Datensatz weist fehler auf. Puffer zu kurz, lang oder fehlerhafte Laengenabgabe.
		RecordBufferOverflow,       // Der Datensatz geht ueber die Grenze des uebergebenen Puffers hinweg.
		FileRead,                   // Es ist ein Dateifehler beim Lesen der Daten aufgetreten.
		FileEmpty,                  // In der Datei liegt kein Datensatz vor.
		// ErrorState der Dateiverarbeitung
		CrcFail,                    // CRC von Daten falsch.
		Format,                     // Bereichsfehler bei eingelesenen Daten usw..
		NoData,                     // Keine Daten beim laden einer Datei.
		LoadDisplayData,            // Laden der Anzeigedaten gescheitert.
		LoadDeviceText,             // Laden der Anzeigetexte gescheitert.
		DeviceUnknown,              // Unbekannter Gertetyp der Gertedatei.
		ProgramOverflow,            // Programmspeicher Ueberlauf.
		RecordMake,                 // Erzeugen eines IntelHex Datensatzes ist fehlgeschlagen.
		CryptFixFail,               // Verschluesselungswert ungueltig.
		CryptNone,                  // Entschluesselungsversuch ohne vorherige Verschluesselung.
		CryptIs,                    // Verschluesselungsversuch obwohl bereits verschluesselt.
		IniKey,                     // Schluesselwert aus Inidatei konnte nicht geladen werden.
		IniKeyValue,                // Schluesselwert aus Inidatei liegt ausserhalb des gueltigen Berreiches.
		Size,                       // Groesse der moeglichen Datenmenge ueberschritten.
		Version,                    // Version der Geraetedatei ist aktueller als erwartet.
		RecordEndMake,              // Daten oder Endzeile fuer Hexdatei konnte nicht erstellt werden.
		HeadCrc,                    // Der CRC fuer den Dateikopf schlug fehl.
		FirmwareVersionLength,      // Zeichenkette fuer die Firmwareversion ist zu kurz (< 8) oder zu lang (> 32).
		ProgramCrc,                 // Der CRC fuer den Programm-Flash schlug fehl.
		FunctionUnavailable,        // Funktion wird nicht unterstuetzt.
		LoadRawFirmware,            // Laden der Rohen Firmwaredaten gescheitert.
		VersionFile,                // Keine gueltige Dateiversionsangabe vorhanden.
		AsText,                     // Fehlertext steht ueber string zur Verfuegung.
		FileFormat,                 // Das Dateiformat ist nicht wie erwartet Intel-Hex.
		LineFormat,                 // Das Format einer Datenzeile ist falsch.
		AllocMem,                   // Speicherallokation konnte nicht durchgefuehrt werden.
		NoFileMatch,                // Es wurde keine zutreffende Geraetedatei in diesem Archiv gefunden.
		SerializeFileData           // Verarbeitung der Dateidaten ist gescheitert.
	};

	enum EnumRecordType 
	{
		// Standard - Datensatztypen bei Intel Hex 32
		TypeDAT                  = 0x00, // Data record
		TypeEOF                  = 0x01, // End-of-file record
		TypeESA                  = 0x02, // Extended Segment Address Record
		TypeSSA                  = 0x03, // Start Segment Address Record
		TypeELA                  = 0x04, // Extended Linear Address Record
		TypeSLA                  = 0x05, // Start Linear Address Record
		// Datei der Anzeigedaten
		TypeDisplayFont          = 0x21,
		TypeDisplayGrafic        = 0x22,
		// Datei der Geraetedaten
		TypeDeviceHeader         = 0x30,
		// Datei der Textdaten
		TypeTextData             = 0x51,
		// Datei der Touchkeydaten
		TypeTouchkeyData         = 0x61,
		// Datei der Displayfarben
		TypeDisplayColorData     = 0x71,
		// Datei der WlanConfig
		TypeWlanConfigData       = 0x81,
		// Datei der HardwareConfig
		TypeHardwareConfigData   = 0x91,
        // Datei der Hauptmenues
        TypeMainMenuData         = 0xA1
	};

	// helper
public:
    virtual void reset();

    virtual bool load( const std::string &filePath );
    virtual bool save( const std::string &filePath, bool append = false );

    inline std::vector<unsigned char> content()
    {
        return _buffer;
    }
    inline unsigned int sizeData()
    {
        return _buffer.size();
    }
    unsigned int firstAddress()
    {
        return _baseAddress;
    }
    unsigned int lastAddress()
    {
        return _baseAddress + _buffer.size();
    }
    unsigned int checksum32bit();

protected:
    std::string _filePath;
    inline std::string filePath() const
    { 
        return _filePath;
    }
    void setFilePath( std::string filePath )
    {
        _filePath = filePath;
    }

protected:
    EnumRecordType _recordType;
    inline EnumRecordType recordType() const
    { 
        return _recordType;
    }
    void setRecordType( EnumRecordType type )
    {
        _recordType = type;
    }

protected:
    enum EnumAddressMode
    {
        Unknown,
        Segment,
        Linear
    } _addressMode;
    struct {
        unsigned short csRegister;
        unsigned short ipRegister;
        bool           exists;
    } _startSegmentAddress;

    struct {
        unsigned int eipRegister;
        bool         exists;
    } _startLinearAddress;

protected:
    unsigned int _baseAddress;
    std::vector<bool> _bufferFillBytes;
    std::vector<unsigned char> _buffer;

    const unsigned char _fillByte;
    const unsigned int _blockSize;

public:
    inline unsigned int blockSize()
    {
        return _blockSize;
    }
    inline unsigned char fillByte()
    {
        return _fillByte;
    }
    inline EnumAddressMode addressMode()
    {
        return _addressMode;
    }
    inline void setAddressMode( EnumAddressMode mode )
    {
        _addressMode = mode;
    }

public:
	void UINT8ToHex( unsigned char val, char *data );
	unsigned char HexToUINT8( const char *data );

	void UINT16ToHex( unsigned short val, char *data );
	unsigned short HexToUINT16( const char *data );

	int MakeCrc( char *data );
	int CheckCrc( const char *data );

public:
	char *PrepareDataLine( char *data, int length );
	int CheckContent( const char *fileName );
	unsigned char GetCount( const char *data );
	unsigned short GetAddress( const char *data );
	unsigned char GetType( const char *data );
	unsigned char GetData( const char *data, unsigned char *dest );
	unsigned char GetData8( const char *data, unsigned char num );
	unsigned short GetData16( const char *data, unsigned char num );
	virtual int MakeTypedDataLine( unsigned short address, unsigned char type, unsigned char *buf, unsigned char length, char *data );
	virtual int MakeDataLine( unsigned short address, unsigned char *buf, unsigned char length, char *data );
	virtual int MakeEndLine( char *data );
	int CheckFile( const char *fileName );

public:
	inline unsigned char maxDataSize()
	{
		return _maxDataSize;
	}

private:
	const unsigned char _maxDataSize;

	// Fehlerbehandlung
protected:
	virtual void resetError();
public:
	virtual std::string errorString( EnumErrorState state );

protected:
	inline void setErrorState( EnumErrorState state )
	{
		_errorState = state;
		if ( state != AsText ) // Fehlertext zur Nummer zuweisen.
			_errorText = errorString( _errorState );
	}
public:
	inline EnumErrorState errorState()
	{
		return _errorState;
	}

protected:
	inline void setErrorText( const std::string &text )
	{
		_errorState = AsText; // Eigenen Fehlertext gesetzt.
		_errorText = text;
	}
public:
	inline std::string errorText()
	{
		return _errorText;
	}

protected:
	inline void setErrorDetail( const std::string &detail )
	{
		_errorDetail = detail;
	}
public:
	inline std::string errorDetail()
	{
		return _errorDetail;
	}

private:
	static EnumErrorState _errorState;
	static std::string _errorText;
	static std::string _errorDetail;
};

#endif
